// COSC 4P76 - Project
// Michael Gircys (mg12vp, 5269550)
// 2016-04-21

package ec.app.COSC4P76_Proj;

import java.awt.Color;
import java.util.ArrayList;
import java.util.List;
import org.apache.commons.lang.ArrayUtils;

import ec.app.IEAS.IEASIndividual;

// In addition to the individual's rendering,
// additional features are extracted and stored 
@SuppressWarnings("serial")
public class IEASAnalyzedIndividual extends IEASIndividual
{
	
	public double[][] Features = new double[FeatureRegionType.values().length][FeatureType.values().length];
	public double GetFeature(FeatureRegionType region, FeatureType feature)
	{
		return Features[region.ordinal()][feature.ordinal()]; 
	}
	public void SetFeature(FeatureRegionType region, FeatureType feature, double value)
	{
		Features[region.ordinal()][feature.ordinal()] = value;
	}
	
	
	double[][][] Rendering_HSV;
	double[][][] Rendering_YCbCr;
	int[][]      Rendering_LBP;
	int[][]      Rendering_RGB_Quantized;
	
	// Listing of regions indexed by FeatureRegionType
	// Specified by {X0,Y0,X1,Y1} ( recall: 0,0 is top left )
	public static double[][] FeatureRegions = new double[][]
		{
			new double[]{ 0.00, 0.00, 1.00, 1.00 },	// Full
			new double[]{ 0.00, 0.00, 0.50, 0.50 },	// UpperLeft
			new double[]{ 0.50, 0.00, 1.00, 0.50 },	// UpperRight
			new double[]{ 0.00, 0.50, 0.50, 1.00 },	// BottomLeft
			new double[]{ 0.50, 0.50, 1.00, 1.00 },	// BottomRight
			new double[]{ 0.25, 0.25, 0.75, 0.75 },	// Center
		};
	
	
	public void Analyze()
	{	
		Rendering_HSV           = new double[Rendering.getWidth()][Rendering.getHeight()][3];
		Rendering_YCbCr         = new double[Rendering.getWidth()][Rendering.getHeight()][3];
		Rendering_LBP           = new int[Rendering.getWidth()][Rendering.getHeight()];
		Rendering_RGB_Quantized = new int[Rendering.getWidth()][Rendering.getHeight()];
		
		// Calculate intermediates for full region
		for(int x = 0; x < Rendering.getWidth();  x++)
		for(int y = 0; y < Rendering.getHeight(); y++)
		{	
			int argb = Rendering.getRGB(x, y);
			Color pColor = new Color(argb);
			int r = pColor.getRed();
			int g = pColor.getGreen();
			int b = pColor.getBlue();
			
			// HSV Colour Values
			float[] colour_hsv = new float[3];
			Color.RGBtoHSB(pColor.getRed(), pColor.getGreen(), pColor.getBlue(), colour_hsv);
			Rendering_HSV[x][y][0] = colour_hsv[0];
			Rendering_HSV[x][y][1] = colour_hsv[1];
			Rendering_HSV[x][y][2] = colour_hsv[2];
			
			// YCbCr Colour Values
			// <http://www.compression.ru/download/articles/color_space/ch03.pdf>
			Rendering_YCbCr[x][y][0] = ( 0.213*r)+(0.715*g)+(0.072*b);
			Rendering_YCbCr[x][y][1] = (-0.213*r)+(0.715*g)+(0.072*b);
			Rendering_YCbCr[x][y][2] = ( 0.213*r)+(0.715*g)+(0.072*b);
			
			// LBP Texture Values
			Rendering_LBP[x][y] = ComputeLBP(x,y);
			
			// Quantize RGB into 512 bins (2^9 --> 3bits/channel)
			Rendering_RGB_Quantized[x][y] = 0
					+ ( ((r>>5)&0x7) << 6 )
					+ ( ((g>>5)&0x7) << 3 )
					+ ( ((b>>5)&0x7) << 0 );
		}
		
		// Extract features for regions
		for( FeatureRegionType region : FeatureRegionType.values() )
		{
			int x0 = (int)Math.floor( FeatureRegions[region.ordinal()][0] * Rendering.getWidth()  );
			int y0 = (int)Math.floor( FeatureRegions[region.ordinal()][1] * Rendering.getHeight() );
			int x1 = (int)Math.floor( FeatureRegions[region.ordinal()][2] * Rendering.getWidth()  );
			int y1 = (int)Math.floor( FeatureRegions[region.ordinal()][3] * Rendering.getHeight() );
			
			int datapoints = (int)(Math.abs(x1-x0)*Math.abs(y1-y0));
			double[] dataset_h         = new double[datapoints];
			double[] dataset_s         = new double[datapoints];
			double[] dataset_v         = new double[datapoints];
			double[] dataset_y         = new double[datapoints];
			int[]    dataset_lbp       = new int[datapoints];
			int[]    dataset_rgb_quant = new int[datapoints];
			
			int idx = 0;
			for(int x = x0; x < x1; x++)
			for(int y = y0; y < y1; y++)
			{
				dataset_h[idx]         = Rendering_HSV[x][y][0];
				dataset_s[idx]         = Rendering_HSV[x][y][1];
				dataset_v[idx]         = Rendering_HSV[x][y][2];
				dataset_y[idx]         = Rendering_YCbCr[x][y][0];
				dataset_lbp[idx]       = Rendering_LBP[x][y];
				dataset_rgb_quant[idx] = Rendering_RGB_Quantized[x][y];
				idx++;
			}
			
			Moments moments_h = new Moments(dataset_h);
			Moments moments_s = new Moments(dataset_h);
			Moments moments_v = new Moments(dataset_h);
			
			SetFeature( region, FeatureType.Colour_H_Mean,   moments_h.Mean   );
			SetFeature( region, FeatureType.Colour_H_StdDev, moments_h.StdDev );
			SetFeature( region, FeatureType.Colour_H_Skew,   moments_h.Skew   );
			SetFeature( region, FeatureType.Colour_S_Mean,   moments_s.Mean   );
			SetFeature( region, FeatureType.Colour_S_StdDev, moments_s.StdDev );
			SetFeature( region, FeatureType.Colour_S_Skew,   moments_s.Skew   );
			SetFeature( region, FeatureType.Colour_V_Mean,   moments_v.Mean   );
			SetFeature( region, FeatureType.Colour_V_StdDev, moments_v.StdDev );
			SetFeature( region, FeatureType.Colour_V_Skew,   moments_v.Skew   );
			
			SetFeature( region, FeatureType.Benford_Lightness, ComputeBenfordDelta(dataset_y));
			
			int LBP_Bin_Count = 4;
			int LBP_Bin_Range = 256/LBP_Bin_Count;
			List<Double> LBP_Bin_1 = new ArrayList<Double>();
			List<Double> LBP_Bin_2 = new ArrayList<Double>();
			List<Double> LBP_Bin_3 = new ArrayList<Double>();
			List<Double> LBP_Bin_4 = new ArrayList<Double>();			
			for(int point : dataset_lbp)
			{
				if      (point < (1 * LBP_Bin_Range)) LBP_Bin_1.add((double)point);
				else if (point < (2 * LBP_Bin_Range)) LBP_Bin_2.add((double)point);
				else if (point < (3 * LBP_Bin_Range)) LBP_Bin_3.add((double)point);
				else                                  LBP_Bin_4.add((double)point);
			}
			Moments moments_lbp1 = new Moments(ArrayUtils.toPrimitive(LBP_Bin_1.toArray(new Double[0]), 0.0));
			Moments moments_lbp2 = new Moments(ArrayUtils.toPrimitive(LBP_Bin_2.toArray(new Double[0]), 0.0));
			Moments moments_lbp3 = new Moments(ArrayUtils.toPrimitive(LBP_Bin_3.toArray(new Double[0]), 0.0));
			Moments moments_lbp4 = new Moments(ArrayUtils.toPrimitive(LBP_Bin_4.toArray(new Double[0]), 0.0));
			
			SetFeature( region, FeatureType.LBP_Bin1_Mean,   moments_lbp1.Mean   );
			SetFeature( region, FeatureType.LBP_Bin1_StdDev, moments_lbp1.StdDev );
			SetFeature( region, FeatureType.LBP_Bin2_Mean,   moments_lbp2.Mean   );
			SetFeature( region, FeatureType.LBP_Bin2_StdDev, moments_lbp2.StdDev );
			SetFeature( region, FeatureType.LBP_Bin3_Mean,   moments_lbp3.Mean   );
			SetFeature( region, FeatureType.LBP_Bin3_StdDev, moments_lbp3.StdDev );
			SetFeature( region, FeatureType.LBP_Bin4_Mean,   moments_lbp4.Mean   );
			SetFeature( region, FeatureType.LBP_Bin4_StdDev, moments_lbp4.StdDev );
			
			SetFeature( region, FeatureType.Complexity_H,    ComputeComplexity(dataset_h, 360));
			SetFeature( region, FeatureType.Complexity_S,    ComputeComplexity(dataset_s, 100));
			SetFeature( region, FeatureType.Complexity_V,    ComputeComplexity(dataset_v, 100));
			SetFeature( region, FeatureType.Complexity_RGB,  ComputeComplexity(dataset_rgb_quant, 512));
			SetFeature( region, FeatureType.Complexity_Y709, ComputeComplexity(dataset_y, 256));
			
			// TODO: Eventually, maybe.
			SetFeature( region, FeatureType.Order_FractalCompression, 0.0);
			SetFeature( region, FeatureType.Metric_MachadoCardoso,    0.0);
		}
		
		return;
	}
	
	
	private static final int[][] LBP_Order = new int[][]
			{	
				{-1,-1}, { 0,-1}, { 1,-1},
				{-1, 0},          { 1,-1},
				{-1, 1}, { 0, 1}, { 1, 1}
			};
	private static final double LBP_Epsilon = 0.0001; // < (1/256)
	private int ComputeLBP( int x, int y )
	{
		int LBP = 0x000000;
		for(int i = 0; i < LBP_Order.length; i++)
		{	
			LBP = LBP << 1;
			int posX = x + LBP_Order[i][0];
			int posY = y + LBP_Order[i][1];
			
			if (posX < 0 || posX >= Rendering.getWidth())  continue;
			if (posY < 0 || posY >= Rendering.getHeight()) continue;
			
			double PosVal = Rendering_YCbCr[posX][posY][0];
			double CenVal = Rendering_YCbCr[x][y][0];
			if( (PosVal-CenVal) >=  LBP_Epsilon  )
				LBP += 0x01;
		}
		return LBP;
	}
	
	private static final double[] Benford_Distribution = new double[]
			{ 0.000, 0.301, 0.176, 0.125, 0.097, 0.079, 0.067, 0.058, 0.051, 0.046 };
	private double ComputeBenfordDelta(double[] vals)
	{
		int validSamples = 0;
		double[] firstDigitFrequencies = new double[10];
		
		for(int i = 0; i < vals.length; i++)
		{	
			double val = Math.abs(vals[i]);
			if(val==0 || Double.isInfinite(val) || Double.isNaN(val)) continue;
			String numStr = String.valueOf(val);
			for(int c = 0; c < numStr.length(); c++)
			{
				char digitStr = numStr.charAt(c);
				if( digitStr != '0' && Character.isDigit(digitStr) )
				{
					int digit = Integer.valueOf(""+digitStr);
					firstDigitFrequencies[digit]++;
					validSamples++;
				}
			}
		}
		
		if(validSamples == 0) return 0;
		
		double measure = 0.0;
		for(int i = 0; i <= 9; i++)
		{
			firstDigitFrequencies[i] /= (double)validSamples;			
			// Li et al. don't take absolute, but that seems wrong.
			measure += Math.abs(firstDigitFrequencies[i] - Benford_Distribution[i]);
		}
		
		return measure;
	}
	
	private double ComputeComplexity(double[] vals, int bins)
	{
		int[] intVals = new int[vals.length];
		for(int i = 0; i < vals.length; i++)
		{
			intVals[i] = (int)Math.floor(vals[i] * bins);
			// Clamp
			if (intVals[i] >= bins) intVals[i] = bins;
			if (intVals[i] <     0) intVals[i] = 0;
		}
		return ComputeComplexity(intVals, bins);
	}
	private double ComputeComplexity(int[] vals, int bins)
	{
		double[] frequencies = new double[bins];
		for(int i = 0; i < vals.length; i++)
			if(vals[i] >= 0 && vals[i] < bins)
				frequencies[vals[i]]++;
		
		double entropy = 0.0;
		for(int i = 0; i < bins; i++)
		{
			frequencies[i] /= vals.length;
			if(frequencies[i]==0) continue;
			entropy += frequencies[i] * Math.log(frequencies[i]); 
		}
		entropy = -entropy; //* vals.length;
		
		return entropy;
	}
	
}
