// COSC 4P76 - Project
// Michael Gircys (mg12vp, 5269550)
// 2016-04-21

package ec.app.COSC4P76_Proj;

import ec.*;
import ec.util.*;

import java.io.IOException;
import java.io.File;

// A statistics class which outputs Learning System 
// classification accuracy and confusion matrices
@SuppressWarnings("serial")
public class IEASLearningStats extends Statistics 
{
	
	// log file parameter
	public static final String P_STATISTICS_FILE = "file";
  
	// The Statistics' log stream
	public int statisticslog = 0;  // stdout
		
  	public void setup(final EvolutionState state, final Parameter base)
    {
  		super.setup(state,base);
              
  		File statisticsFile = state.parameters.getFile(base.push(P_STATISTICS_FILE),null);
  		
  		if (silentFile)
  			statisticslog = Output.NO_LOGS;
  		else if (statisticsFile!=null)
        {
  			try
  			{
  				statisticslog = state.output.addLog(statisticsFile, true, false);
            }
  			catch (IOException i)
            {
              	state.output.fatal("An IOException occurred while trying to create the log " + statisticsFile + ":\n" + i);
            }
        }
  		else 
  			state.output.warning("No statistics file specified, printing to stdout at end.", base.push(P_STATISTICS_FILE));
  		
    }

	public void postInitializationStatistics(final EvolutionState state)
 	{
		super.postInitializationStatistics(state);
 	}

	public void postEvaluationStatistics(final EvolutionState state)
	{
		super.postEvaluationStatistics(state);
		
		int gen = state.generation;
		IEASLearning prob = (IEASLearning)(state.evaluator.p_problem);
		LearningSystem learnSys = prob.Learner;
		
		state.output.print(""   + gen , statisticslog);
		
		if(learnSys.Evaluation_ANN == null || learnSys.Evaluation_Tree == null)
		{
			state.output.println("", statisticslog);
			return;
		}
		
		state.output.print(", " + (prob.isUserDeferringRating() ? 0 : 1) , statisticslog);
		
		state.output.print(", " + learnSys.Evaluation_Tree.pctCorrect() , statisticslog);
		state.output.print(", " + learnSys.Evaluation_Tree.pctIncorrect() , statisticslog);
		state.output.print(", " + learnSys.Evaluation_Tree.pctUnclassified() , statisticslog);
		double[][] confusion_tree = learnSys.Evaluation_Tree.confusionMatrix();
		for(int i = 0; i < confusion_tree.length; i++)
		for(int j = 0; j < confusion_tree[i].length; j++)
			state.output.print(", " + confusion_tree[i][j] , statisticslog);
		
		state.output.print(", " + learnSys.Evaluation_ANN.pctCorrect() , statisticslog);
		state.output.print(", " + learnSys.Evaluation_ANN.pctIncorrect() , statisticslog);
		state.output.print(", " + learnSys.Evaluation_ANN.pctUnclassified() , statisticslog);
		double[][] confusion_ann = learnSys.Evaluation_ANN.confusionMatrix();
		for(int i = 0; i < confusion_ann.length; i++)
		for(int j = 0; j < confusion_ann[i].length; j++)
			state.output.print(", " + confusion_ann[i][j] , statisticslog);
		
		state.output.println("", statisticslog);
	}

	protected void bypassFinalStatistics(EvolutionState state, int result)
	{
		super.finalStatistics(state, result); 
	}

	public void finalStatistics(final EvolutionState state, final int result)
    {
		super.finalStatistics(state,result);		
		// If required, write some final stats
    }
	
}
