// COSC 4P76 - Project
// Michael Gircys (mg12vp, 5269550)
// 2016-04-21

package ec.app.COSC4P76_Proj;

import java.awt.*;
import java.awt.event.*;
import java.util.Collection;
import java.util.LinkedList;
import javax.swing.*;
import ec.EvolutionState;
import ec.Individual;
import ec.Subpopulation;
import ec.gp.GPSpecies;
import ec.app.IEAS.*;

// Based on the previous IEASWindow, the IEASLearningWindow provides an additional
// button to allow for user rating deferal, and tracks whether or not the current 
// generation has been deferred.
// Heavy code reuse was required as the singleton pattern proves difficult to extend.
@SuppressWarnings("serial")
public class IEASLearningWindow extends JFrame
{

	// Singleton.
	private static IEASLearningWindow _Instance = null;
	public static IEASLearningWindow Instance()
	{
		if (_Instance == null) _Instance = new IEASLearningWindow();
		return _Instance;
	}
	private IEASLearningWindow()
	{	
		SetupWindow();
		try {
	        UIManager.setLookAndFeel(UIManager.getSystemLookAndFeelClassName());
	    }catch(Exception ex) {
	        ex.printStackTrace();
	    }
	}
	
	// User exit status
	protected static boolean UserExited = false;
	public static boolean hasUserExited(){ return UserExited; }
	// User deferred evaluation
	protected static boolean UserDeferred = false;
	public static boolean hasUserDeferred(){ return UserDeferred; }
	public static void setUserDeferred(boolean isDeferred){ UserDeferred = isDeferred; }
	
	
	public JPanel ItemPanel;	// Individuals
	public JPanel SidePanel;	// State Config.	
	protected void SetupWindow()
	{	
		this.setSize(1024, 768);
		this.setExtendedState(JFrame.MAXIMIZED_BOTH);
		this.setDefaultCloseOperation(JFrame.DO_NOTHING_ON_CLOSE);
		this.addWindowListener( new WindowAdapter(){
			@Override
		    public void windowClosing(java.awt.event.WindowEvent windowEvent) 
			{
				JFrame wind = IEASLearningWindow.Instance();
		        if (JOptionPane.showConfirmDialog(wind, 
		            "This will terminate the system.\nAre you sure you want to exit?", "Really Closing?", 
		            JOptionPane.YES_NO_OPTION, JOptionPane.QUESTION_MESSAGE) 
		            == JOptionPane.YES_OPTION)
		        {
		            UserExited = true;
		            wind.setVisible(false);
		            wind.dispose();
		            UserFinished();
		        }
		    }
		});
		
		this.getContentPane().setLayout(new GridBagLayout());
		this.getContentPane().setBackground(Color.BLACK);
		
		SetupSidePanel();
		SetupItemPanel();
	}
	public void SetupItemPanel()
	{
		JPanel p = ItemPanel = new JPanel();
		GridBagConstraints c = new GridBagConstraints();
		p.setLayout( new WrapLayout(FlowLayout.CENTER, 10, 10) );
		p.setBackground(Color.DARK_GRAY);
		c = new GridBagConstraints();
		c.gridx = 0;
		c.fill = GridBagConstraints.BOTH;
		c.anchor = GridBagConstraints.LINE_START;
		c.weightx = 1.0;
		c.weighty = 1.0;
		
		JScrollPane jsp = new JScrollPane(p, JScrollPane.VERTICAL_SCROLLBAR_AS_NEEDED, JScrollPane.HORIZONTAL_SCROLLBAR_NEVER);
		this.getContentPane().add(jsp,c);
		
		// Items to be rated will be populated on a separate call.
	}
	public void SetupSidePanel()
	{
		JPanel FullSidePanel = new JPanel();
		FullSidePanel.setLayout(new BorderLayout());
		FullSidePanel.setMinimumSize(  new Dimension(300,0));
		FullSidePanel.setPreferredSize(new Dimension(300,0));
		GridBagConstraints c = new GridBagConstraints();
		c.gridx = 1;
		c.fill = GridBagConstraints.BOTH;
		c.anchor = GridBagConstraints.LINE_END;
		c.weightx = 0.0;
		c.weighty = 1.0;
		this.getContentPane().add(FullSidePanel, c);
		//----------------------------
		
		// Side panel hook
		JPanel p = SidePanel = new JPanel();
		p.setLayout(new GridBagLayout());
		JScrollPane jsp = new JScrollPane(p, JScrollPane.VERTICAL_SCROLLBAR_AS_NEEDED, JScrollPane.HORIZONTAL_SCROLLBAR_NEVER);
		FullSidePanel.add(jsp, BorderLayout.CENTER);
		
		// Other configuration settings on the side panel to be populated on state bind.
		
		JPanel ButtonContainer = new JPanel();
		ButtonContainer.setLayout(new GridBagLayout());
		// Continue Button
		JButton b_cont = new JButton();
		c = new GridBagConstraints();
		c.gridy = 0;
		c.fill = GridBagConstraints.HORIZONTAL;
		b_cont.setText("Continue");
		b_cont.setMinimumSize(  new Dimension(250,100));
		b_cont.setPreferredSize(new Dimension(300,100));
		b_cont.addActionListener(new ActionListener(){
			@Override public void actionPerformed(ActionEvent arg0){ IEASLearningWindow.Instance().UserFinished(); }
		});
		ButtonContainer.add(b_cont,c);
		// Defer Button
		JButton b_auto = new JButton();
		c = new GridBagConstraints();
		c.gridy = 1;
		c.fill = GridBagConstraints.HORIZONTAL;
		b_auto.setText("Automatic Evaluation");
		b_auto.setMinimumSize(  new Dimension(250,100));
		b_auto.setPreferredSize(new Dimension(300,100));
		b_auto.addActionListener(new ActionListener(){
			@Override public void actionPerformed(ActionEvent arg0){ IEASLearningWindow.Instance().UserDeferred(); }
		});
		ButtonContainer.add(b_auto,c);
		//Attach
		FullSidePanel.add(ButtonContainer, BorderLayout.PAGE_END);
	}
	
	
	protected EvolutionState BoundState;
	// The collections should be the same as the items in the ItemPanel/SidePanel, but maybe
	// we'll be adding other types of items in there later or something.
	protected Collection<IEASImageRatingComponent> RatingBoxes = new LinkedList<IEASImageRatingComponent>();
	protected Collection<IEASPopSettingComponent>  PopSettings = new LinkedList<IEASPopSettingComponent>();
	// Bind EvolutionState, so we can fiddle with the GP configuration and stuff.
	public void BindEvolutionState(EvolutionState state)
	{	
		BoundState = state;
		// -------------------
		
		SidePanel.removeAll();
		PopSettings.clear();
		
		GridBagConstraints c = new GridBagConstraints();
		c.insets = new Insets(5,5,5,5);
		int cur_grid_y = -1;
		JLabel l;
		
		// Setting Label
		l = new JLabel();
		l.setText("Setting");
		c.anchor = GridBagConstraints.PAGE_START;
		c.gridy = ++cur_grid_y;
		c.gridx = 0;
		SidePanel.add(l,c);
		// Value Label
		l = new JLabel();
		l.setText("Value");
		c.anchor = GridBagConstraints.PAGE_START;
		c.gridy = cur_grid_y;
		c.gridx = 1;
		SidePanel.add(l,c);

		// list population evolver pipeline settings.
		c = new GridBagConstraints();
		c.gridx = 0;
		c.gridwidth = 2;
		c.anchor = GridBagConstraints.CENTER;
		c.fill = GridBagConstraints.HORIZONTAL;
		c.weightx = 1.0;
		for(int s = 0; s < BoundState.population.subpops.length; s++)
		{	
			Subpopulation subpop = BoundState.population.subpops[s];
			IEASPopSettingComponent psc = new IEASPopSettingComponent((GPSpecies)(subpop.species), s, BoundState);
			PopSettings.add(psc);
			c.gridy = ++cur_grid_y;
			SidePanel.add(psc,c);
		}

	}
	protected void UpdateStateSettings()
	{
		// Set items in the BoundState.
		for(IEASPopSettingComponent psc : PopSettings)
			psc.Confirm();
	}
	public void RefreshIndividuals()
	{
		ItemPanel.removeAll();
		RatingBoxes.clear();
		
		for( Subpopulation subpop : BoundState.population.subpops )
		for( Individual indv : subpop.individuals )
		{
			// It better be an IEASIndividual
			IEASImageRatingComponent rater = new IEASImageRatingComponent( (IEASIndividual)indv, BoundState );
			ItemPanel.add( rater );
			RatingBoxes.add( rater );
		}
		this.revalidate();
		this.repaint();
	}
	protected void UpdateFitnessValues()
	{
		for(IEASImageRatingComponent rating : RatingBoxes)
			rating.Confirm();
	}
	
	
	// Wait for user to rank individuals.
	// Go ahead, relax, dip your feet in the water.
	protected Object waitingPool = new Object();	
	public void WaitForUser()
	{	
		synchronized (waitingPool) {
			HideWaitingMessage();
			try {
				waitingPool.wait();
			} catch (InterruptedException e) {
				e.printStackTrace();
			}
		}
	}	
	public void UserFinished()
	{	
		UserDeferred = false;
		synchronized (waitingPool) {
			ShowWaitingMessage();
			UpdateFitnessValues();
			UpdateStateSettings();
			waitingPool.notifyAll();
		}
	}
	public void UserDeferred()
	{
		UserDeferred = true;
		synchronized (waitingPool) {
			ShowWaitingMessage();
			UpdateFitnessValues();
			UpdateStateSettings();
			waitingPool.notifyAll();
		}
	}
	
	
	// Just put a grey overlay on top of things to show that we're waiting.
	protected void ShowWaitingMessage() 
	{  
		JPanel glass = (JPanel)(this.getGlassPane());
		Color glassColor = new Color(200,200,200,255);
		glass.setOpaque(true);
        glass.setBackground(glassColor);
        glass.setVisible(true);
	}
	protected void HideWaitingMessage() 
	{  
		JPanel glass = (JPanel)(this.getGlassPane());
		glass.setVisible(false);
	}
	
}
