// COSC 4P76 - Project
// Michael Gircys (mg12vp, 5269550)
// 2016-04-21

package ec.app.COSC4P76_Proj;

import ec.EvolutionState;
import ec.Individual;
import ec.gp.koza.KozaFitness;
import ec.util.Parameter;

// Methods for automatically assigning an objective fitness score to images 
public class IEASSimpleAutomaticFitness 
{

	public int Defer_For      = 0;
	public int Defer_After    = 0;
	
	// Specifys which scheme (formula) we will be using for fitness
	public int Fitness_Scheme = 0;
	
	// If we're on the final generation; show the window so we can see how it did
	public boolean Show_Final_Gen = false;
	
	public void setup(final EvolutionState state, final Parameter base)
	{
		Fitness_Scheme = state.parameters.getIntWithDefault(base.push("scheme"),      null, 0);
		Defer_For      = state.parameters.getIntWithDefault(base.push("defer_for"),   null, 0);
		Defer_After    = state.parameters.getIntWithDefault(base.push("defer_after"), null, 0);
		Show_Final_Gen = state.parameters.getBoolean(base.push("showfinal"),          null, false);
	}
	
	// Sets the fitness of an individual using a configured fixed fitness scheme
	public void SetFitness(EvolutionState state, Individual ind) 
	{
		IEASAnalyzedIndividual indv = (IEASAnalyzedIndividual)ind;
		
		int hits = GetRating(state,indv);  
		
		KozaFitness f = ((KozaFitness)ind.fitness);
        f.hits = hits;
        double fitness = ((double)(hits-1)) / 4.0;
        f.setStandardizedFitness(state, 1.0 - fitness); // 0.0==>Good, 1.0==>Bad
	}

	// Returns rating in domain [1,2,3,4,5]
	public int GetRating(EvolutionState state, IEASAnalyzedIndividual ind)
	{
		switch(Fitness_Scheme)
		{
			case 1: return GetRating_1(state,ind);
			case 2: return GetRating_2(state,ind);
			case 3: return GetRating_3(state,ind);
			default: return 1; // The whole world must be awful if we have an invalid scheme!
		}
	}
	// A simplistic fixed fitness scheme.
	// Fitness is determined through an absolute measure (not relative to other individuals).
	// We are looking for images with a full window having a blue mean hue.
	protected int GetRating_1(EvolutionState state, IEASAnalyzedIndividual ind)
	{
		double full_hue = ind.GetFeature(FeatureRegionType.Full, FeatureType.Colour_H_Mean);
		double hue_blue = (240/360);
		double dist = Math.abs(hue_blue-full_hue);
		if(dist > 0.5) dist -= 0.5; // Hue is periodic.
		dist = (dist*8)+1.0; // [0.0, 0.5] --> [1.0,5.0]
		return (int)Math.round(dist);
	}
	// Sadly, further schemes were not required, since the simplest scheme proved difficult.
	protected int GetRating_2(EvolutionState state, IEASAnalyzedIndividual ind)
	{
		return 1;
	}
	protected int GetRating_3(EvolutionState state, IEASAnalyzedIndividual ind)
	{
		return 1;
	}
	
	
	// Track state for whether or not we should classify a generation using
	// the learning system or fixed fitness scheme
	protected boolean Deferring_State      = false;
	protected int GenerationsDeferred      = 0;
	protected int GenerationsSinceDeferral = 0;
	
	// Check if the current generation should be classified using the
	// learning system (instead of our fixed fitness scheme)
	// when using a fixed fitness scheme for testing
	public boolean IsDeferal()
	{
		return Deferring_State;
	}
	
	// Notify that we have completed a generation, and deferal state 
	// should potential be updated based on configuration 
	public void NextGeneration()
	{
		if(Deferring_State)
		{
			GenerationsDeferred++;
			if (GenerationsDeferred >= Defer_For)
			{
				GenerationsDeferred = 0;
				GenerationsSinceDeferral = 0;
				if(Defer_After > 0)
					Deferring_State = false;
			}
		}
		else
		{
			GenerationsSinceDeferral++;
			if (GenerationsSinceDeferral >= Defer_After)
			{
				GenerationsDeferred = 0;
				GenerationsSinceDeferral = 0;
				if(Defer_For > 0)
					Deferring_State = true;
			}
		}
	}
	
	
	// Test harness
	public static void main(String[] args)
	{
		IEASSimpleAutomaticFitness AutoFitness = new IEASSimpleAutomaticFitness();
		AutoFitness.Defer_For = 2;
		AutoFitness.Defer_After = 1;
		for(int i = 0; i < 50; i++)
		{
			System.out.println("Gen " + i + ": " + (AutoFitness.IsDeferal()?"Deferral":"Active") );
			AutoFitness.NextGeneration();
		}
	}
	
	
}
