package ec.app.COSC5P71_Proj;

import ec.EvolutionState;
import ec.Problem;
import ec.gp.ADFStack;
import ec.gp.ERC;
import ec.gp.GPData;
import ec.gp.GPIndividual;
import ec.gp.GPNode;
import ec.util.Code;
import ec.util.DecodeReturn;


@SuppressWarnings("serial")
public abstract class Func_Ephemeral extends ERC
{
	public double value;
	
	@Override public String name()   { return "E"; }
	@Override public String encode() { return Code.encode(value); }
	
	@Override
	public String toStringForHumans()
	{
		return name() + "[" + String.valueOf(value) + "]";
	}
	
	@Override
	public boolean nodeEquals(GPNode node) 
	{
		if(node instanceof Func_Ephemeral) return false;
		return Math.abs(((Func_Ephemeral)node).value - this.value) < 1.0E-15 ;
	}
	
	// From the manual.
	@Override
	public boolean decode(DecodeReturn dret) 
	{
		int    pos 	= dret.pos;
		String data = dret.data;
		Code.decode(dret);
		if (dret.type != DecodeReturn.T_DOUBLE)
		{ 
			dret.data = data; 
			dret.pos = pos; 
			return false; 
		}
		value = dret.d;
		return true;
	}
	
	public void eval(
	    	final EvolutionState state,
	        final int            thread,
	        final GPData         input,
	        final ADFStack       stack,
	        final GPIndividual   individual,
	        final Problem        problem
    )
    {   
        DoubleData rd = ((DoubleData)(input));
        rd.x = value;
    }	    
	
	
	
	// Eph_1 --> [0.0, 1.0]
	public static class Eph_1 extends Func_Ephemeral
	{	
		@Override
		public void resetNode(EvolutionState state, int thread) 
		{
			value = state.random[thread].nextDouble(true,true);
		}
		@Override
		public void mutateERC(EvolutionState state, int thread) 
		{
			double v;
			do v = value + state.random[thread].nextGaussian() * 0.01;
			while( v < 0.0 || v > 1.0 );
			value = v;
		}
	}
	
	// Eph_10 --> [0.0, 10.0]
	public static class Eph_10 extends Func_Ephemeral
	{	
		@Override
		public void resetNode(EvolutionState state, int thread) 
		{
			value = state.random[thread].nextDouble(true,true) * 10.0;
		}
		@Override
		public void mutateERC(EvolutionState state, int thread) 
		{
			double v;
			do v = value + state.random[thread].nextGaussian() * 0.1;
			while( v < 0.0 || v > 10.0 );
			value = v;
		}
	}
	
	// Eph_100 --> [0.0, 100.0]
	public static class Eph_100 extends Func_Ephemeral
	{	
		@Override
		public void resetNode(EvolutionState state, int thread) 
		{
			value = state.random[thread].nextDouble(true,true) * 100.0;
		}
		@Override
		public void mutateERC(EvolutionState state, int thread) 
		{
			double v;
			do v = value + state.random[thread].nextGaussian() * 1.0;
			while( v < 0.0 || v > 100.0 );
			value = v;
		}
	}

}
