package ec.app.COSC5P71_Proj;

// Interactive Evolutionary Art System

import java.awt.image.BufferedImage;

import ec.EvolutionState;
import ec.Individual;
import ec.gp.GPProblem;
import ec.gp.koza.KozaFitness;
import ec.util.Parameter;


@SuppressWarnings("serial")
public class IEAS extends GPProblem implements IEASProblemForm 
{
	
	protected static final int Default_Image_Size = 128;
	
	public double Current_Pos_X;					// The current position for which to evaluate the trees.
	public double Current_Pos_Y;
	
	
	// We expect a [0.0,1.0] range. Make it [0,255] (fairly). 
	private int FitToChannelRange(double v)
	{
		v = Math.min(Math.max( 0.0, v), 1.0 );
		return (int)Math.ceil( v * 255.0 );
	}
	
	
	@Override
	public void setup(final EvolutionState state, final Parameter base)
    {
        super.setup(state, base);
        //Parameter params = base;
        
        // verify our input is the right class (or subclasses from it)
        if (!(input instanceof DoubleData))
            state.output.fatal("GPData class must subclass from " + DoubleData.class, base.push(P_DATA), null);

        // Not much else to check or setup.
        state.output.systemMessage("Problem initialized.");
    }
	
	// This is done across multiple threads.
	// It's been relegated to produce the image phenotype.
	// User evaluation is managed by the IEASEvaluator, which calls our userEvaluate() method.
	@Override
	public void evaluate(EvolutionState state, Individual ind, int subpopulation, int threadnum) 
	{
		if(ind.evaluated) return;
		
		if(ind instanceof IEASIndividual)
		{
			IEASIndividual indv = (IEASIndividual)ind;
			indv.Rendering = RenderImage(state, indv, subpopulation, threadnum, Default_Image_Size);
		}
		
		// Default fitness.
        KozaFitness f = ((KozaFitness)ind.fitness);
        f.hits = 0;
        f.setStandardizedFitness(state, 1.0); // 0.0==>Good, 1.0==>Bad
        
        ind.evaluated = true;
	}
	
	
	// Separated, and with parameterized size, to allow for calls from the window
	// (ie. user wants to look at image details, or save a huge image)
	public BufferedImage RenderImage(EvolutionState state, IEASIndividual ind, int subpop, int threadnum, int size)
	{
		final int Image_Size_X = size;
		final int Image_Size_Y = size;
		// Percentage of image that a pixel takes up
		final double Image_Step_X = 1.0 / ((double) Image_Size_X);
		final double Image_Step_Y = 1.0 / ((double) Image_Size_Y);
		
		DoubleData input = (DoubleData)(this.input);
		BufferedImage outImage = new BufferedImage(Image_Size_X, Image_Size_Y, BufferedImage.TYPE_INT_RGB);
		
		for(int y = 0; y < Image_Size_Y; y++)
		for(int x = 0; x < Image_Size_X; x++)
		{
			int r, g, b, rgb = 0;
			Current_Pos_X = Image_Step_X * x;
			Current_Pos_Y = Image_Step_Y * y;
			
			if( ind.trees.length <3 )
			{	
				// Oh ... uh, okay then? I guess grayscale?
				ind.trees[0].child.eval(state,threadnum,input,stack,ind,this);
				r = g = b = FitToChannelRange(input.x);
			}
			else
			{
				ind.trees[0].child.eval(state,threadnum,input,stack,ind,this);
				r = FitToChannelRange(input.x);
				ind.trees[1].child.eval(state,threadnum,input,stack,ind,this);
				g = FitToChannelRange(input.x);
				ind.trees[2].child.eval(state,threadnum,input,stack,ind,this);
				b = FitToChannelRange(input.x);
			}
			
			rgb = (r<<16) + (g<<8) + (b);
			outImage.setRGB(x, y, rgb);
		}
		
		return outImage;
	}
	
	
	public void prepareToUserEvaluate(final EvolutionState state)
	{	
		IEASWindow.Instance().BindEvolutionState(state);
		IEASWindow.Instance().setVisible(true);
	}
	public void userEvaluate(final EvolutionState state)
	{
		IEASWindow.Instance().RefreshIndividuals();
		IEASWindow.Instance().WaitForUser();
	}	
	public void userEvaluationFinished(final EvolutionState state)
	{
		// Nah, we don't need to scale the user ratings or anything
		// since we're using tournament selection.
	}	
	public boolean userExited()
	{
		return IEASWindow.hasUserExited();
	}
}
