package ec.app.COSC5P71_Proj;

import java.awt.*;
import java.util.*;

import javax.swing.*;

import ec.BreedingSource;
import ec.EvolutionState;
import ec.Prototype;
import ec.gp.GPSpecies;
import ec.simple.SimpleBreeder;


@SuppressWarnings("serial")
public class IEASPopSettingComponent extends JPanel 
{
	
	protected int BoundSubpopNumber;
	protected GPSpecies BoundSpecies;
	protected EvolutionState BoundState;
	
	public IEASPopSettingComponent(GPSpecies species, int subpopNum, EvolutionState state)
	{
		BoundSpecies = species;
		BoundState = state;
		BoundSubpopNumber = subpopNum;
		initialize();
	}
	
	
	protected JSpinner ControlElitism;
	protected Dictionary<String, JComponent> PipeProbControls = new Hashtable<String, JComponent>();
	
	protected void initialize()
	{
		this.setLayout(new GridBagLayout());
		
		int cur_grid_y = -1;
		JLabel l;
		
		// label constraints.		
		GridBagConstraints lc = new GridBagConstraints();
		lc.gridx = 0;
		lc.anchor = GridBagConstraints.LINE_START;
		lc.insets = new Insets( 3,3,3,3 );
		lc.fill = GridBagConstraints.HORIZONTAL;
		// Control constraints.
		GridBagConstraints cc = new GridBagConstraints();
		cc.gridx = 1;
		cc.anchor = GridBagConstraints.LINE_START;
		cc.insets = new Insets( 3,3,3,3 );
		cc.fill = GridBagConstraints.HORIZONTAL;
		
		// Population ID label
		l = new JLabel("Population ["+ BoundSubpopNumber +"] :");
		lc.gridy = cc.gridy = ++cur_grid_y;
		this.add(l,lc);
		this.add(new JLabel(""),cc);
		
		// Pipelines
		// Probabilites have been normalized, and have been adjusted such that they
		// represent a chunk of linear space from the last last pipeline probability.
		// (ie. a random in [0.0,1.0] has the correct probability of hitting a given pipeline) 
		// We need to undo this first.
		double PreviousNormalizedProbPoint = 0.0;
		for(BreedingSource bs : BoundSpecies.pipe_prototype.sources)
		{
			String pipeName = ((Prototype)bs).defaultBase().toString();
			int    pipeProb = (int)Math.round((bs.probability-PreviousNormalizedProbPoint)*100);
			PreviousNormalizedProbPoint = bs.probability;
			JComponent pipeCont = new JSpinner( new SpinnerNumberModel(
						 pipeProb		// current
						,0				// min
						,100			// max
						,1				// step
					) );
			
			l = new JLabel("% " + pipeName);
			lc.gridy = cc.gridy = ++cur_grid_y;
			this.add(l,lc);
			this.add(pipeCont,cc);
			
			PipeProbControls.put(pipeName, pipeCont);
		}
		
		// Breeder Elitism
		l = new JLabel("Elitism");
		lc.gridy = cc.gridy = ++cur_grid_y;
		ControlElitism = new JSpinner( new SpinnerNumberModel(
					 ((SimpleBreeder)(BoundState.breeder)).elite[BoundSubpopNumber]			// current
					,0 																		// min
					,BoundState.population.subpops[BoundSubpopNumber].individuals.length	// max
					,1																		// step
				) );
		this.add(l,lc);
		this.add(ControlElitism,cc);
	}
	
	
	public void Confirm()
	{
		// Pipeline Probabilities
		for(BreedingSource bs : BoundSpecies.pipe_prototype.sources)
		{
			JSpinner cont = (JSpinner)(PipeProbControls.get(bs.defaultBase().toString()));
			int val = (Integer)(cont.getValue());
			bs.probability = ((double)val) / 100.0;
		}
		// Redo the probability conversion 
		ec.BreedingSource.setupProbabilities(BoundSpecies.pipe_prototype.sources);
				
		// Breeder Elitism
		((SimpleBreeder)BoundState.breeder).elite[BoundSubpopNumber] = (Integer)(ControlElitism.getValue());
	}
	
}
