package ec.app.IEAS;
import ec.*;
import ec.gp.*;

@SuppressWarnings("serial")
public abstract class Func_Conditional extends GPNode 
{
	
	// -- double input, double output -----------------------------------------

	// If C1 Greater-Than C2
	public static class IfGT extends Func_Conditional
	{
	    public String toString()         { return "IfGT"; }	    
	    public int    expectedChildren() { return 4; }	
	    public void eval(
	    	final EvolutionState state,
	        final int            thread,
	        final GPData         input,
	        final ADFStack       stack,
	        final GPIndividual   individual,
	        final Problem        problem
	    )
	    {
	        double c1, c2, val;
	        MultiData rd = ((MultiData)(input));
	
	        children[0].eval(state,thread,input,stack,individual,problem);
	        c1 = rd.GetD();
	        children[1].eval(state,thread,input,stack,individual,problem);
			c2 = rd.GetD();
			
			children[ (c1>c2)? 2 : 3 ].eval(state,thread,input,stack,individual,problem);
			val = rd.GetD();
			
	        rd.SetTo(  val  );
	    }	    
	}
	
	// If C1 Greater-Than 0
	public static class IfG0 extends Func_Conditional
	{
	    public String toString()         { return "IfG0"; }	    
	    public int    expectedChildren() { return 3; }	
	    public void eval(
	    	final EvolutionState state,
	        final int            thread,
	        final GPData         input,
	        final ADFStack       stack,
	        final GPIndividual   individual,
	        final Problem        problem
	    )
	    {
	        double cond, val;
	        MultiData rd = ((MultiData)(input));
	
	        children[0].eval(state,thread,input,stack,individual,problem);
	        cond = rd.GetD();
	        
			children[ (cond >= 0.0)? 1 : 2 ].eval(state,thread,input,stack,individual,problem);
			val = rd.GetD();
			
	        rd.SetTo(  val  );
	    }
	}
	
	// If Unsigned C1 Greater-Than 0 (C-style, approximation for Double)
	public static class If extends Func_Conditional
	{
	    public String toString()         { return "If"; }	    
	    public int    expectedChildren() { return 3; }	
	    public void eval(
	    	final EvolutionState state,
	        final int            thread,
	        final GPData         input,
	        final ADFStack       stack,
	        final GPIndividual   individual,
	        final Problem        problem
	    )
	    {
	    	MultiData rd = ((MultiData)(input));
	    	
	    	final double Delta = 1.0E-14;
	        double cond, val;
	        boolean condMet;
	        
	        children[0].eval(state,thread,input,stack,individual,problem);
	        cond = rd.GetD();
	        
	        condMet = Math.abs(cond) < Delta;
	        
			children[ condMet? 1 : 2 ].eval(state,thread,input,stack,individual,problem);
			val = rd.GetD();
			
	        rd.SetTo(  val  );
	    }	    
	}
	
	
	
	// -- double input, binary output -----------------------------------------	
	
	public static class bIfGT extends Func_Conditional
	{
	    public String toString()         { return "IfGT"; }	    
	    public int    expectedChildren() { return 2;      }	
	    public void eval( final EvolutionState state, final int thread, final GPData input, final ADFStack stack, final GPIndividual individual, final Problem problem )
	    {
	    	MultiData rd = ((MultiData)(input));
	    	children[0].eval(state,thread,input,stack,individual,problem);
	        double c1 = rd.GetD();
	        children[1].eval(state,thread,input,stack,individual,problem);
	        double c2 = rd.GetD();	        
	        rd.SetTo(  c1 > c2  );
	    }	    
	}	
	public static class bIfG0 extends Func_Conditional
	{
	    public String toString()         { return "IfG0"; }	    
	    public int    expectedChildren() { return 1;      }	
	    public void eval( final EvolutionState state, final int thread, final GPData input, final ADFStack stack, final GPIndividual individual, final Problem problem )
	    {
	    	MultiData rd = ((MultiData)(input));
	        children[0].eval(state,thread,input,stack,individual,problem);
	        double c1 = rd.GetD();
	        rd.SetTo(  c1 > 0.0  );
	    }	    
	}
	public static class bIf extends Func_Conditional
	{
	    public String toString()         { return "If"; }	    
	    public int    expectedChildren() { return 1;    }	
	    public void eval( final EvolutionState state, final int thread, final GPData input, final ADFStack stack, final GPIndividual individual, final Problem problem )
	    {
	    	MultiData rd = ((MultiData)(input));
	    	final double Delta = 1.0E-14;
	        double cond;
	        
	        children[0].eval(state,thread,input,stack,individual,problem);
	        cond = rd.GetD();
	        
	        rd.SetTo(  Math.abs(cond) < Delta  );
	    }
	}
	
	
	
	// -- binary inputs -------------------------------------------------------
	
	public static class IF_double extends Func_Conditional
	{
		public String toString()         { return "If"; }
	    public int    expectedChildren() { return 3;    }
	    public void eval( final EvolutionState state, final int thread, final GPData input, final ADFStack stack, final GPIndividual individual, final Problem problem )
	    {
	    	MultiData rd = ((MultiData)(input));
	        children[0].eval(state,thread,input,stack,individual,problem);
	        boolean cond = rd.GetB();
	        children[ cond? 1 : 2 ].eval(state,thread,input,stack,individual,problem);
	        double val = rd.GetD();
	        rd.SetTo(  val  );
	    }
	}

	public static class IF_texture extends Func_Conditional
	{
		public String toString()         { return "If"; }
	    public int    expectedChildren() { return 3;    }
	    public void eval( final EvolutionState state, final int thread, final GPData input, final ADFStack stack, final GPIndividual individual, final Problem problem )
	    {
	    	MultiData rd = ((MultiData)(input));
	        children[0].eval(state,thread,input,stack,individual,problem);
	        boolean cond = rd.GetB();
	        children[ cond? 1 : 2 ].eval(state,thread,input,stack,individual,problem);
	        java.awt.image.BufferedImage val = rd.GetT();
	        rd.SetTo(  val  );
	    }
	}
	
	public static class IF_vector extends Func_Conditional
	{
		public String toString()         { return "If"; }
	    public int    expectedChildren() { return 3;    }
	    public void eval( final EvolutionState state, final int thread, final GPData input, final ADFStack stack, final GPIndividual individual, final Problem problem )
	    {
	    	MultiData rd = ((MultiData)(input));
	        children[0].eval(state,thread,input,stack,individual,problem);
	        boolean cond = rd.GetB();
	        children[ cond? 1 : 2 ].eval(state,thread,input,stack,individual,problem);
	        double[] val = rd.GetS();
	        rd.SetTo(  val  );
	    }
	}
	
	
}
	