package ec.app.IEAS;
import ec.*;
import ec.gp.*;

@SuppressWarnings("serial")
public abstract class Func_Logic extends GPNode 
{

	public static class TRUE extends Func_Logic
	{
		public String toString()         { return "TRUE"; }	    
	    public int    expectedChildren() { return 0;      }	
	    public void eval(
	    	final EvolutionState state,
	        final int            thread,
	        final GPData         input,
	        final ADFStack       stack,
	        final GPIndividual   individual,
	        final Problem        problem
	    )
	    {
	        MultiData rd = ((MultiData)(input));
	        rd.SetTo(  true  );
	    }	    
	}
	public static class FALSE extends Func_Logic
	{
		public String toString()         { return "FALSE"; }	    
	    public int    expectedChildren() { return 0;       }	
	    public void eval(
	    	final EvolutionState state,
	        final int            thread,
	        final GPData         input,
	        final ADFStack       stack,
	        final GPIndividual   individual,
	        final Problem        problem
	    )
	    {
	        MultiData rd = ((MultiData)(input));
	        rd.SetTo(  false  );
	    }	    
	}
	
	public static class NOT extends Func_Logic
	{
		public String toString()         { return "NOT"; }	    
	    public int    expectedChildren() { return 1;     }	
	    public void eval(
	    	final EvolutionState state,
	        final int            thread,
	        final GPData         input,
	        final ADFStack       stack,
	        final GPIndividual   individual,
	        final Problem        problem
	    )
	    {
	        boolean val;
	        MultiData rd = ((MultiData)(input));
	
	        children[0].eval(state,thread,input,stack,individual,problem);
	        val = rd.GetB();
	        
	        rd.SetTo(  !val  );
	    }	    
	}
	
	public static class AND extends Func_Logic
	{
		public String toString()         { return "AND"; }	    
	    public int    expectedChildren() { return 2;     }	
	    public void eval(
	    	final EvolutionState state,
	        final int            thread,
	        final GPData         input,
	        final ADFStack       stack,
	        final GPIndividual   individual,
	        final Problem        problem
	    )
	    {
	        boolean v1, v2;
	        MultiData rd = ((MultiData)(input));
	
	        children[0].eval(state,thread,input,stack,individual,problem);
	        v1 = rd.GetB();
	        children[1].eval(state,thread,input,stack,individual,problem);
	        v2 = rd.GetB();
	        
	        rd.SetTo(  v1 & v2  );
	    }	    
	}
	
	public static class OR extends Func_Logic
	{
		public String toString()         { return "OR"; }	    
	    public int    expectedChildren() { return 2;     }	
	    public void eval(
	    	final EvolutionState state,
	        final int            thread,
	        final GPData         input,
	        final ADFStack       stack,
	        final GPIndividual   individual,
	        final Problem        problem
	    )
	    {
	        boolean v1, v2;
	        MultiData rd = ((MultiData)(input));
	
	        children[0].eval(state,thread,input,stack,individual,problem);
	        v1 = rd.GetB();
	        children[1].eval(state,thread,input,stack,individual,problem);
	        v2 = rd.GetB();
	        
	        rd.SetTo(  v1 | v2  );
	    }	    
	}
	
	public static class NAND extends Func_Logic
	{
		public String toString()         { return "NAND"; }	    
	    public int    expectedChildren() { return 2;     }	
	    public void eval(
	    	final EvolutionState state,
	        final int            thread,
	        final GPData         input,
	        final ADFStack       stack,
	        final GPIndividual   individual,
	        final Problem        problem
	    )
	    {
	        boolean v1, v2;
	        MultiData rd = ((MultiData)(input));
	
	        children[0].eval(state,thread,input,stack,individual,problem);
	        v1 = rd.GetB();
	        children[1].eval(state,thread,input,stack,individual,problem);
	        v2 = rd.GetB();
	        
	        rd.SetTo(  !(v1&v2)  );
	    }	    
	}
	
	public static class XOR extends Func_Logic
	{
		public String toString()         { return "XOR"; }	    
	    public int    expectedChildren() { return 2;     }	
	    public void eval(
	    	final EvolutionState state,
	        final int            thread,
	        final GPData         input,
	        final ADFStack       stack,
	        final GPIndividual   individual,
	        final Problem        problem
	    )
	    {
	        boolean v1, v2;
	        MultiData rd = ((MultiData)(input));
	
	        children[0].eval(state,thread,input,stack,individual,problem);
	        v1 = rd.GetB();
	        children[1].eval(state,thread,input,stack,individual,problem);
	        v2 = rd.GetB();
	        
	        rd.SetTo(  (v1|v2)&(!(v1&v2))  );
	    }	    
	}
	
	public static class IMPLIES extends Func_Logic
	{
		public String toString()         { return "IMPL"; }	    
	    public int    expectedChildren() { return 2;     }	
	    public void eval(
	    	final EvolutionState state,
	        final int            thread,
	        final GPData         input,
	        final ADFStack       stack,
	        final GPIndividual   individual,
	        final Problem        problem
	    )
	    {
	        boolean v1, v2;
	        MultiData rd = ((MultiData)(input));
	
	        children[0].eval(state,thread,input,stack,individual,problem);
	        v1 = rd.GetB();
	        children[1].eval(state,thread,input,stack,individual,problem);
	        v2 = rd.GetB();
	        
	        rd.SetTo(  v1 ? v2 : true  );
	    }	    
	}
	
	public static class IF extends Func_Logic
	{
		public String toString()         { return "IF"; }	    
	    public int    expectedChildren() { return 3;    }	
	    public void eval(
	    	final EvolutionState state,
	        final int            thread,
	        final GPData         input,
	        final ADFStack       stack,
	        final GPIndividual   individual,
	        final Problem        problem
	    )
	    {
	        boolean v, cond;
	        MultiData rd = ((MultiData)(input));
	        
	        children[0].eval(state,thread,input,stack,individual,problem);
	        cond = rd.GetB();
	        children[ cond? 1 : 2 ].eval(state,thread,input,stack,individual,problem);
	        v = rd.GetB();
	        
	        rd.SetTo(  v  );
	    }	    
	}
	
}
