package ec.app.IEAS;

import java.awt.*;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.WindowAdapter;
import java.awt.image.BufferedImage;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.PrintWriter;
import javax.imageio.ImageIO;
import javax.swing.*;
import javax.swing.tree.DefaultMutableTreeNode;
import javax.swing.tree.DefaultTreeModel;

import ec.EvolutionState;
import ec.gp.GPNode;
import ec.gp.GPTree;


@SuppressWarnings("serial")
public class IEASImageDetailWindow extends JFrame
{
	// Potentially singleton - user may want 1 window or several.
	protected static IEASImageDetailWindow LastWindow;
	public static boolean UseSingleWindow = true;
	public static IEASImageDetailWindow GetWindow(IEASIndividual indv, EvolutionState state)
	{
		if(LastWindow == null || !UseSingleWindow) 
			LastWindow = new IEASImageDetailWindow();
		LastWindow.BindImage(indv, state);
		return LastWindow;
	}
	
	
	protected final int ImageSize = 800; // Detailed display is a 800x800 image.
	protected final String[] ChannelNames = new String[]{ "R", "G", "B" };
	
	protected EvolutionState BoundState;
	protected IEASIndividual BoundIndividual; 

	protected IEASImageDetailWindow()
	{
		initialize();
	}
	protected IEASImageDetailWindow(IEASIndividual indv, EvolutionState state) 
	{ 
		initialize();
		BindImage(indv, state); 
	}
	
	public void BindImage(IEASIndividual indv, EvolutionState state) 
	{
		BoundIndividual = indv;
		BoundState = state;
		UpdateBoundDisplay();
	}
	
	
	BufferedImage BiggerSourceImage;
	JLabel ImageIcon;
	JTextField ImageSaveSize;
	JRadioButton rbImageFilter_Norm;
	JRadioButton rbImageFilter_ChanR;
	JRadioButton rbImageFilter_ChanG;
	JRadioButton rbImageFilter_ChanB;
	JRadioButton rbImageFilter_Multi;
	JCheckBox cbImageFilter_Gray;
	JTree TreeView;
	
	protected void initialize()
	{
		this.setSize(1024, 768);
		this.setExtendedState(JFrame.MAXIMIZED_BOTH);
		this.setDefaultCloseOperation(JFrame.DO_NOTHING_ON_CLOSE);
		this.addWindowListener( new WindowAdapter(){
			@Override public void windowClosing(java.awt.event.WindowEvent windowEvent) { windowEvent.getWindow().setVisible(false); }
		});		
		this.getContentPane().setLayout(new GridBagLayout());
		this.getContentPane().setBackground(Color.BLACK);
		
		GridBagConstraints c; 	// panel constraints
		GridBagConstraints ic; 	// item constraints
		
		// Image ==========================================================
		
		JPanel pImage = new JPanel();
		pImage.setBackground(Color.WHITE);
		c = new GridBagConstraints();
		c.gridx = 0;
		c.gridy = 0;
		c.anchor = GridBagConstraints.CENTER;
		c.fill = GridBagConstraints.BOTH;
		c.weightx = 1.0;
		c.weighty = 1.0;
		this.getContentPane().add(pImage, c);
		
		pImage.setLayout(new GridBagLayout());
		ic = new GridBagConstraints();
		ic.anchor = GridBagConstraints.CENTER;
		ic.fill = GridBagConstraints.NONE;
		ic.insets = new Insets( 5,5,5,5 );
		ImageIcon = new JLabel();
		pImage.add(ImageIcon,ic);
		
		// Bigger image display handled on bind.
		
		// Tree display ==========================================================
		
		JPanel pTree = new JPanel();
		c = new GridBagConstraints();
		c.gridx = 1;
		c.gridy = 0;
		c.anchor = GridBagConstraints.FIRST_LINE_START;
		c.fill = GridBagConstraints.BOTH;
		c.weightx = 1.0;
		c.weighty = 1.0;
		this.getContentPane().add(pTree, c);
		
		TreeView = new JTree();
		pTree.setLayout(new GridBagLayout());
		TreeView.setBorder( BorderFactory.createEmptyBorder(10,10,10,10) );
		ic = new GridBagConstraints();
		ic.anchor = GridBagConstraints.CENTER;
		ic.fill = GridBagConstraints.BOTH;
		ic.weightx = 1.0;
		ic.weighty = 1.0;
		JScrollPane TreeScroll = new JScrollPane(TreeView);
		pTree.add(TreeScroll,ic);
		
		// Tree specifics handled on bind.
		
		// Image options ==========================================================
		
		JPanel pImageOpts = new JPanel();
		c = new GridBagConstraints();
		c.gridx = 0;
		c.gridy = 1;
		c.anchor = GridBagConstraints.FIRST_LINE_START;
		c.fill = GridBagConstraints.HORIZONTAL;
		c.weightx = 1.0;
		c.weighty = 0.0;
		this.getContentPane().add(pImageOpts, c);
		
		rbImageFilter_Norm  = new JRadioButton("Normal", true);
		rbImageFilter_ChanR = new JRadioButton("Channel[R]");
		rbImageFilter_ChanG = new JRadioButton("Channel[G]");
		rbImageFilter_ChanB = new JRadioButton("Channel[B]");
		rbImageFilter_Multi = new JRadioButton("Multiple");
		cbImageFilter_Gray  = new JCheckBox("Channels Show As Grayscale.");
		// Associate RadioButtons to single group.
		ButtonGroup bg = new ButtonGroup();
		bg.add(rbImageFilter_Norm );
		bg.add(rbImageFilter_ChanR);
		bg.add(rbImageFilter_ChanG);
		bg.add(rbImageFilter_ChanB);
		bg.add(rbImageFilter_Multi);
		
		pImageOpts.setLayout(new GridBagLayout());
		int cur_grid_x = -1;
		ic = new GridBagConstraints();		
		ic.insets = new Insets( 10,10,5,10 );
		ic.gridy = 0;
		
		ic.gridx = ++cur_grid_x;
		pImageOpts.add(rbImageFilter_Norm , ic);
		ic.gridx = ++cur_grid_x;
		pImageOpts.add(rbImageFilter_Multi, ic);
		ic.gridx = ++cur_grid_x;
		pImageOpts.add(rbImageFilter_ChanR, ic);
		ic.gridx = ++cur_grid_x;
		pImageOpts.add(rbImageFilter_ChanG, ic);
		ic.gridx = ++cur_grid_x;
		pImageOpts.add(rbImageFilter_ChanB, ic);
		
		ic.gridy = 1;
		ic.gridx = 1; // start _after_ the first.
		ic.gridwidth = cur_grid_x; // expand to the end of the previous items.
		ic.insets = new Insets( 0,0,5,0 );
		pImageOpts.add(cbImageFilter_Gray , ic);
		
		// On click, update display.
		ActionListener ImageOptActionListener = new ActionListener(){
			@Override public void actionPerformed(ActionEvent arg0) { UpdateBoundImage(); }
		};
		rbImageFilter_Norm .addActionListener(ImageOptActionListener);
		rbImageFilter_Multi.addActionListener(ImageOptActionListener);
		rbImageFilter_ChanR.addActionListener(ImageOptActionListener);
		rbImageFilter_ChanG.addActionListener(ImageOptActionListener);
		rbImageFilter_ChanB.addActionListener(ImageOptActionListener);
		cbImageFilter_Gray .addActionListener(ImageOptActionListener);
		
		// Save Options ==========================================================
		
		JPanel pSaveOpts = new JPanel();
		c = new GridBagConstraints();
		c.gridx = 1;
		c.gridy = 1;
		c.anchor = GridBagConstraints.FIRST_LINE_START;
		c.fill = GridBagConstraints.BOTH;
		c.weightx = 1.0;
		c.weighty = 0.0;
		this.getContentPane().add(pSaveOpts, c);
		
		JButton bSaveImage = new JButton("Save Image");
		bSaveImage.addActionListener( new ActionListener(){
			@Override public void actionPerformed(ActionEvent arg0) { SaveImage(); }
		});
		
		JButton bSaveTrees = new JButton("Save Trees");
		bSaveTrees.addActionListener( new ActionListener(){
			@Override public void actionPerformed(ActionEvent arg0) { SaveTrees(); }
		});
		
		ImageSaveSize = new JTextField();
		ImageSaveSize.setText(String.valueOf(ImageSize));
		ImageSaveSize.setColumns(5);
		
		pSaveOpts.setLayout(new FlowLayout( FlowLayout.CENTER, 5, 5 ));
		pSaveOpts.add(new JLabel("Image Size: "));
		pSaveOpts.add(ImageSaveSize);
		pSaveOpts.add(bSaveImage);
		pSaveOpts.add(bSaveTrees);
		
	}	
	
	protected void UpdateBoundDisplay()
	{
		// Render the image in a suitably large size.
		BiggerSourceImage = ((IEAS)(BoundState.evaluator.p_problem)).RenderImage(BoundState, BoundIndividual, 0, 0, ImageSize);
		UpdateBoundImage();
		
		//TreeView
		DefaultMutableTreeNode root = new DefaultMutableTreeNode("Individual");
		for(int i = 0; i < BoundIndividual.trees.length; i++)
		{
			String chanName = "?";
			if(BoundIndividual.trees.length == 3) 
				chanName = ChannelNames[i];
				
			DefaultMutableTreeNode chanHeader = new DefaultMutableTreeNode("Channel["+chanName+"]");
			DefaultMutableTreeNode chanNodes  = GPNode_to_JTree(BoundIndividual.trees[i].child);
			
			chanHeader.add(chanNodes);
			root.add(chanHeader);
		}
		TreeView.setModel( new DefaultTreeModel(root) );
		for (int i = 0; i < TreeView.getRowCount(); i++)
		    TreeView.expandRow(i);
		
	}
	protected void UpdateBoundImage()
	{
		if(BiggerSourceImage==null) return;
		
		// Copy the bigger image, 'cause we might want to fiddle with the channels or stuff.
		BufferedImage DisplayedImage = FilterImage(BiggerSourceImage);
		
		// Display new bigger and adjusted image
		Dimension picSizing = new Dimension(DisplayedImage.getWidth(), DisplayedImage.getHeight());
		ImageIcon.setIcon(new ImageIcon( DisplayedImage ));
		ImageIcon.setBorder( BorderFactory.createLineBorder(Color.BLACK,1) );
		ImageIcon.setMinimumSize( picSizing );
		ImageIcon.setMaximumSize( picSizing );
		ImageIcon.setPreferredSize(picSizing);
		ImageIcon.revalidate();
	}
	
	// Recursively convert GPtree to a JTree node representation.
	public DefaultMutableTreeNode GPNode_to_JTree(GPNode node)
	{
		DefaultMutableTreeNode tNode = new DefaultMutableTreeNode( node.toStringForHumans() );
		for(GPNode child : node.children)
			tNode.add(GPNode_to_JTree(child));
		return tNode;
	}
	

	// Isolate channels and otherwise adjust image for display
	public BufferedImage FilterImage(BufferedImage src)
	{
		BufferedImage out = new BufferedImage(
				src.getColorModel(),
				src.copyData(null),
				src.isAlphaPremultiplied(),
				null
			);
		
		// Channel filters.
		if(	rbImageFilter_ChanR.isSelected() ||
			rbImageFilter_ChanG.isSelected() ||
			rbImageFilter_ChanB.isSelected()  )
		{
			int filter = 0xFF << (rbImageFilter_ChanR.isSelected() ? 16 : (rbImageFilter_ChanG.isSelected() ? 8 : 0)); 
			
			for(int y = 0; y < out.getHeight(); y++)
			for(int x = 0; x < out.getWidth();  x++)				
			{
				int newVal = out.getRGB(x,y) & filter;
				
				if(cbImageFilter_Gray.isSelected())
				{	
					newVal = (newVal >> 16) | (newVal >> 8) | newVal;
					newVal = (newVal << 16) | (newVal << 8) | newVal;
				}
				
				out.setRGB(x,y, newVal);
			}
			
			return out;
		}
		
		// Multiple channel overlay
		if(rbImageFilter_Multi.isSelected())
		{
			// Use 2x2 sampling to reduce the image to a quarter of it's size.
			for(int y = 0; y < out.getHeight(); y+=2)
			for(int x = 0; x < out.getWidth() ; x+=2)
			{
				int r,g,b, rgb;
				rgb = r = g = b = 0;
				
				// Average samples.
				for(int offy = 0; offy < 2; offy++)
				for(int offx = 0; offx < 2; offx++)
				{
					rgb = src.getRGB(x+offx,y+offy) & 0xFFFFFF;
					r += ((rgb & 0xFF0000) >> 0x10);
					g += ((rgb & 0x00FF00) >> 0x08);
					b += ((rgb & 0x0000FF) >> 0x00);
				}
				r = (int)Math.round( ((double)r) / 4.0 );
				g = (int)Math.round( ((double)g) / 4.0 );
				b = (int)Math.round( ((double)b) / 4.0 );
				rgb = (r << 16) + (g << 8) + b;

				out.setRGB((x/2),(y/2), rgb);
			}
			
			// Filter the 2-4th quadrants to single channels.
			for(int y = 0; y < out.getHeight()/2; y++)
			for(int x = 0; x < out.getWidth() /2; x++)
			{
				int val = out.getRGB(x,y) & 0xFFFFFF;
				
				int valR = val & 0xFF0000;
				int valG = val & 0x00FF00;
				int valB = val & 0x0000FF;
				
				if(cbImageFilter_Gray.isSelected())
				{
					valR = valR | (valR >> 0x08) | (valR >> 0x10);
					valG = valG | (valG >> 0x08) | (valG << 0x08);
					valB = valB | (valB << 0x08) | (valB << 0x10);
				}
				
				out.setRGB(x+(out.getWidth()/2), y                    , valR);
				out.setRGB(x                   , y+(out.getHeight()/2), valG);
				out.setRGB(x+(out.getWidth()/2), y+(out.getHeight()/2), valB);
			}
		}
		
		// Normal. No adjustments.
		return out;
	}
	
	
	public void SaveImage()
	{
		// Get and validate render size
		String sSize = ImageSaveSize.getText();
		int iSize;
		try {
			iSize = Integer.parseInt(sSize);
		} catch (NumberFormatException e) {
			iSize = 0;
		}
		if(iSize <= 0)
		{
			JOptionPane.showMessageDialog(this, "Exported image size should be a positive integer.");
			return;
		}
		
		if(	iSize > 5000 &&
			JOptionPane.showConfirmDialog(this, 
		        "That's a pretty big render size you've specified there.\nYou sure?", "Correct Parameter?", 
		        JOptionPane.YES_NO_OPTION, JOptionPane.QUESTION_MESSAGE) 
		        != JOptionPane.YES_OPTION)
			return;
		
		// Get output filename
		File outFile;
		JFileChooser fileChooser = new JFileChooser();
		int fileChooserResult = fileChooser.showSaveDialog(this);
		if( fileChooserResult == JFileChooser.APPROVE_OPTION)
			outFile = fileChooser.getSelectedFile();
		else
			return;
		if( !outFile.getName().toUpperCase().endsWith(".PNG") )
			outFile = new File( outFile.toString() + ".png" );
		
		// Generate sized and adjusted image
		BufferedImage tmpImage = ((IEAS)(BoundState.evaluator.p_problem)).RenderImage(
				BoundState, BoundIndividual, 0, 0, iSize);
		BufferedImage outImage = FilterImage(tmpImage);
		
		// OUT!
		try {
			ImageIO.write(outImage, "PNG", outFile);
		} catch (IOException e) {
			e.printStackTrace();
			return;
		}
		
		JOptionPane.showMessageDialog(this, "Image saved.");
	}
	public void SaveTrees()
	{
		// Get output filename
		File outFile;
		JFileChooser fileChooser = new JFileChooser();
		int fileChooserResult = fileChooser.showSaveDialog(this);
		if( fileChooserResult == JFileChooser.APPROVE_OPTION)
			outFile = fileChooser.getSelectedFile();
		else
			return;
		if( !outFile.getName().toUpperCase().endsWith(".TXT") )
			outFile = new File( outFile.toString() + ".txt" );
		
		// Prepare file for output
		PrintWriter writer;
		try {
			 writer = new PrintWriter(outFile);
		} catch (FileNotFoundException e) {
			e.printStackTrace();
			return;
		}
		
		// Write tree to file.
		for( int treeNum = 0; treeNum < BoundIndividual.trees.length; treeNum++)
		{
			// Channel/Tree identifier
			String chan = "?";
			if(BoundIndividual.trees.length == 3)
				chan = ChannelNames[treeNum];
			writer.append(String.format("Channel[%s] %n", chan));
			
			// Tree nodes
			GPTree tree = BoundIndividual.trees[treeNum];
			printGPNodeWithChildren(writer, tree.child, 1);
			
			// Extra space.
			writer.append(String.format("%n"));
		}
		
		writer.flush();
		writer.close();
		
		JOptionPane.showMessageDialog(this, "Tree saved.");
	}

	
	// How is there seriously no method for this that doesn't require a log number?!
	private void printGPNodeWithChildren(PrintWriter writer, GPNode curNode, int tabLevel)
	{	
		for(int t = 0; t < tabLevel; t++) 
			writer.append("\t");
		writer.append(curNode.toStringForHumans());
		writer.append(String.format("%n"));
		
		for(GPNode child : curNode.children)
			printGPNodeWithChildren(writer, child, tabLevel+1);
	}
	
}
