package ec.app.IEAS;

import java.awt.*;
import java.awt.event.*;
import java.util.Collection;
import java.util.LinkedList;
import javax.swing.*;
import ec.EvolutionState;
import ec.Individual;
import ec.Subpopulation;
import ec.gp.GPSpecies;


@SuppressWarnings("serial")
public class IEASWindow extends JFrame
{

	// Singleton.
	private static IEASWindow _Instance = null;
	public static IEASWindow Instance()
	{
		if (_Instance == null) _Instance = new IEASWindow();
		return _Instance;
	}
	private IEASWindow()
	{
		SetupWindow();
		try {
	        UIManager.setLookAndFeel(UIManager.getSystemLookAndFeelClassName());
	    }catch(Exception ex) {
	        ex.printStackTrace();
	    }
	}
	
	// User exit status
	protected static boolean UserExited = false;
	public static boolean hasUserExited()
	{
		return UserExited;
	}
	
	
	public JPanel ItemPanel;	// Individuals
	public JPanel SidePanel;	// State Config.	
	protected void SetupWindow()
	{	
		this.setSize(1024, 768);
		this.setExtendedState(JFrame.MAXIMIZED_BOTH);
		this.setDefaultCloseOperation(JFrame.DO_NOTHING_ON_CLOSE);
		this.addWindowListener( new WindowAdapter(){
			@Override
		    public void windowClosing(java.awt.event.WindowEvent windowEvent) 
			{
				JFrame wind = IEASWindow.Instance();
		        if (JOptionPane.showConfirmDialog(wind, 
		            "This will terminate the system.\nAre you sure you want to exit?", "Really Closing?", 
		            JOptionPane.YES_NO_OPTION, JOptionPane.QUESTION_MESSAGE) 
		            == JOptionPane.YES_OPTION)
		        {
		            UserExited = true;
		            wind.setVisible(false);
		            wind.dispose();
		            UserFinished();
		        }
		    }
		});
		
		this.getContentPane().setLayout(new GridBagLayout());
		this.getContentPane().setBackground(Color.BLACK);
		
		SetupSidePanel();
		SetupItemPanel();
	}
	public void SetupItemPanel()
	{
		JPanel p = ItemPanel = new JPanel();
		GridBagConstraints c = new GridBagConstraints();
		p.setLayout( new WrapLayout(FlowLayout.CENTER, 10, 10) );
		p.setBackground(Color.DARK_GRAY);
		c = new GridBagConstraints();
		c.gridx = 0;
		c.fill = GridBagConstraints.BOTH;
		c.anchor = GridBagConstraints.LINE_START;
		c.weightx = 1.0;
		c.weighty = 1.0;
		
		JScrollPane jsp = new JScrollPane(p, JScrollPane.VERTICAL_SCROLLBAR_AS_NEEDED, JScrollPane.HORIZONTAL_SCROLLBAR_NEVER);
		this.getContentPane().add(jsp,c);
		
		// Items to be rated will be populated on a separate call.
	}
	public void SetupSidePanel()
	{
		JPanel FullSidePanel = new JPanel();
		FullSidePanel.setLayout(new BorderLayout());
		FullSidePanel.setMinimumSize(  new Dimension(300,0));
		FullSidePanel.setPreferredSize(new Dimension(300,0));
		GridBagConstraints c = new GridBagConstraints();
		c.gridx = 1;
		c.fill = GridBagConstraints.BOTH;
		c.anchor = GridBagConstraints.LINE_END;
		c.weightx = 0.0;
		c.weighty = 1.0;
		this.getContentPane().add(FullSidePanel, c);
		//----------------------------
		
		// Side panel hook
		JPanel p = SidePanel = new JPanel();
		p.setLayout(new GridBagLayout());
		JScrollPane jsp = new JScrollPane(p, JScrollPane.VERTICAL_SCROLLBAR_AS_NEEDED, JScrollPane.HORIZONTAL_SCROLLBAR_NEVER);
		FullSidePanel.add(jsp, BorderLayout.CENTER);
		
		// Other configuration settings on the side panel to be populated on state bind.
		
		// Continue Button
		JButton b = new JButton();
		b.setText("Continue");
		b.setMinimumSize(  new Dimension(0,100));
		b.setPreferredSize(new Dimension(0,100));
		b.addActionListener(new ActionListener(){
			@Override public void actionPerformed(ActionEvent arg0){ IEASWindow.Instance().UserFinished(); }
		});
		FullSidePanel.add(b, BorderLayout.PAGE_END);		
	}
	
	
	protected EvolutionState BoundState;
	// The collections should be the same as the items in the ItemPanel/SidePanel, but maybe
	// we'll be adding other types of items in there later or something.
	protected Collection<IEASImageRatingComponent> RatingBoxes = new LinkedList<IEASImageRatingComponent>();
	protected Collection<IEASPopSettingComponent>  PopSettings = new LinkedList<IEASPopSettingComponent>();
	// Bind EvolutionState, so we can fiddle with the GP configuration and stuff.
	public void BindEvolutionState(EvolutionState state)
	{	
		BoundState = state;
		// -------------------
		
		SidePanel.removeAll();
		PopSettings.clear();
		
		GridBagConstraints c = new GridBagConstraints();
		c.insets = new Insets(5,5,5,5);
		int cur_grid_y = -1;
		JLabel l;
		
		// Setting Label
		l = new JLabel();
		l.setText("Setting");
		c.anchor = GridBagConstraints.PAGE_START;
		c.gridy = ++cur_grid_y;
		c.gridx = 0;
		SidePanel.add(l,c);
		// Value Label
		l = new JLabel();
		l.setText("Value");
		c.anchor = GridBagConstraints.PAGE_START;
		c.gridy = cur_grid_y;
		c.gridx = 1;
		SidePanel.add(l,c);

		// list population evolver pipeline settings.
		c = new GridBagConstraints();
		c.gridx = 0;
		c.gridwidth = 2;
		c.anchor = GridBagConstraints.CENTER;
		c.fill = GridBagConstraints.HORIZONTAL;
		c.weightx = 1.0;
		for(int s = 0; s < BoundState.population.subpops.length; s++)
		{	
			Subpopulation subpop = BoundState.population.subpops[s];
			IEASPopSettingComponent psc = new IEASPopSettingComponent((GPSpecies)(subpop.species), s, BoundState);
			PopSettings.add(psc);
			c.gridy = ++cur_grid_y;
			SidePanel.add(psc,c);
		}

	}
	protected void UpdateStateSettings()
	{
		// Set items in the BoundState.
		for(IEASPopSettingComponent psc : PopSettings)
			psc.Confirm();
	}
	public void RefreshIndividuals()
	{
		ItemPanel.removeAll();
		RatingBoxes.clear();
		
		for( Subpopulation subpop : BoundState.population.subpops )
		for( Individual indv : subpop.individuals )
		{
			// It better be an IEASIndividual
			IEASImageRatingComponent rater = new IEASImageRatingComponent( (IEASIndividual)indv, BoundState );
			ItemPanel.add( rater );
			RatingBoxes.add( rater );
		}
		this.revalidate();
		this.repaint();
	}
	protected void UpdateFitnessValues()
	{
		for(IEASImageRatingComponent rating : RatingBoxes)
			rating.Confirm();
	}
	
	
	// Wait for user to rank individuals.
	// Go ahead, relax, dip your feet in the water.
	protected Object waitingPool = new Object();	
	public void WaitForUser()
	{	
		synchronized (waitingPool) {
			HideWaitingMessage();
			try {
				waitingPool.wait();
			} catch (InterruptedException e) {
				e.printStackTrace();
			}
		}
	}	
	public void UserFinished()
	{
		synchronized (waitingPool) {
			ShowWaitingMessage();
			UpdateFitnessValues();
			UpdateStateSettings();
			waitingPool.notifyAll();
		}
	}
	
	
	// Just put a grey overlay on top of things to show that we're waiting.
	protected void ShowWaitingMessage() 
	{  
		JPanel glass = (JPanel)(this.getGlassPane());
		Color glassColor = new Color(200,200,200,255);
		glass.setOpaque(true);
        glass.setBackground(glassColor);
        glass.setVisible(true);
	}
	protected void HideWaitingMessage() 
	{  
		JPanel glass = (JPanel)(this.getGlassPane());
		glass.setVisible(false);
	}

	
	// Test harness.
	public static void main(String[] args)
	{
		EvolutionState es = new ec.simple.SimpleEvolutionState();
		
		es.population = new ec.Population();
		es.population.subpops = new Subpopulation[1];
		es.population.subpops[0] = new Subpopulation();
		es.population.subpops[0].species = new GPSpecies();
		es.population.subpops[0].individuals = new IEASIndividual[56];
		for(int i = 0; i < es.population.subpops[0].individuals.length; i++)
		{
			IEASIndividual iv = new IEASIndividual();
			es.population.subpops[0].individuals[i] = iv;
			
			iv.Rendering = new java.awt.image.BufferedImage(128, 128, java.awt.image.BufferedImage.TYPE_INT_RGB);
			for(int y = 0; y < iv.Rendering.getHeight(); y++)
			for(int x = 0; x < iv.Rendering.getWidth(); x++)
				iv.Rendering.setRGB(x, y, (x<<16) + (y<<8) + (x+y));
		}
		es.population.subpops[0].species.pipe_prototype = new ec.breed.MultiBreedingPipeline();
		es.population.subpops[0].species.pipe_prototype.sources = new ec.BreedingPipeline[2];
		es.population.subpops[0].species.pipe_prototype.sources[0] = new ec.gp.koza.CrossoverPipeline();
		es.population.subpops[0].species.pipe_prototype.sources[0].probability = 0.9;
		es.population.subpops[0].species.pipe_prototype.sources[1] = new ec.gp.breed.MutateERCPipeline();
		es.population.subpops[0].species.pipe_prototype.sources[1].probability = 1.0;
		es.breeder = new ec.simple.SimpleBreeder();
		((ec.simple.SimpleBreeder)(es.breeder)).elite = new int[es.population.subpops.length];
		
		
		Instance().BindEvolutionState(es);
		Instance().setVisible(true);
		Instance().RefreshIndividuals();
	}
	
}
